# Copyright (C) 2012-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the shorten program."""

from .. import fileutil, util


def extract_shn(archive, compression, cmd, verbosity, interactive, outdir):
    """Decompress a SHN archive to a WAV file."""
    cmdlist = [util.shell_quote(cmd)]
    outfile = fileutil.get_single_outfile(outdir, archive, extension=".wav")
    cmdlist.extend(
        ['-x', '-', util.shell_quote(outfile), '<', util.shell_quote(archive)]
    )
    return (cmdlist, {'shell': True})


def create_shn(archive, compression, cmd, verbosity, interactive, filenames):
    """Compress a WAV file to a SHN archive."""
    if len(filenames) > 1:
        raise util.PatoolError("multiple filenames for shorten not supported")
    cmdlist = [util.shell_quote(cmd)]
    cmdlist.extend(
        ['-', util.shell_quote(archive), '<', util.shell_quote(filenames[0])]
    )
    return (cmdlist, {'shell': True})



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the star program"""

from . import ArchiveTest
from .. import needs_program, needs_codec


class TestStar(ArchiveTest):
    """Test class for the star program"""

    program = 'star'

    @needs_program(program)
    def test_star(self):
        """Run archive commands with TAR archives."""
        self.archive_commands('t.tar')
        self.archive_commands('t.cbt')

    @needs_codec(program, 'gzip')
    def test_star_gz(self):
        """Run archive commands with TAR GZIP archives."""
        self.archive_commands('t.tar.gz')
        self.archive_commands('t.tgz')

    @needs_program(program)
    @needs_program('compress')
    def test_star_z(self):
        """Run archive commands with TAR COMPRESS archives."""
        self.archive_commands('t.tar.Z')
        self.archive_commands('t.taz')

    @needs_codec(program, 'bzip2')
    def test_star_bz2(self):
        """Run archive commands with TAR BZIP2 archives."""
        self.archive_commands('t.tar.bz2')
        self.archive_commands('t.tbz2')

    @needs_codec(program, 'lzma')
    def test_star_lzma(self):
        """Run archive commands with TAR LZMA archive."""
        self.archive_commands('t.tar.lzma')

    @needs_program(program)
    @needs_program('lzip')
    def test_star_lzip(self):
        """Run archive commands with TAR LZIP archive."""
        self.archive_commands('t.tar.lz')

    @needs_codec(program, 'xz')
    def test_star_xz(self):
        """Run archive commands with TAR XZ archive."""
        self.archive_commands('t.tar.xz')

    @needs_program('file')
    @needs_program(program)
    def test_star_file(self):
        """Run archive commands with renamed TAR archives."""
        self.archive_commands('t.tar.foo', skip_create=True)
        self.archive_commands('t.cbt.foo', skip_create=True)

    @needs_program('file')
    @needs_codec(program, 'gzip')
    def test_star_gz_file(self):
        """Run archive commands with renamed TAR GZIP archives."""
        self.archive_commands('t.tar.gz.foo', skip_create=True)
        self.archive_commands('t.tgz.foo', skip_create=True)

    @needs_program('file')
    @needs_codec(program, 'compress')
    def test_star_z_file(self):
        """Run archive commands with renamed TAR COMPRESS archives."""
        self.archive_commands('t.tar.Z.foo', skip_create=True)
        self.archive_commands('t.taz.foo', skip_create=True)

    @needs_program('file')
    @needs_codec(program, 'bzip2')
    def test_star_bz2_file(self):
        """Run archive commands with renamed TAR BZIP2 archives."""
        self.archive_commands('t.tar.bz2.foo', skip_create=True)
        self.archive_commands('t.tbz2.foo', skip_create=True)

    # file(1) does not recognize .lzma files
    # @needs_program('file')
    # @needs_codec(program, 'lzma')
    # def test_star_lzma_file(self):
    #    """Run archive commands with renamed TAR LZMA archive."""
    #    self.archive_commands('t.tar.lzma.foo', skip_create=True)

    # file(1) does not recognize .lzip files
    # @needs_program('file')
    # @needs_codec(program, 'lzip')
    # def test_star_lzip_file(self):
    #    """Run archive commands with renamed TAR LZIP archive."""
    #    self.archive_commands('t.tar.lz.foo', skip_create=True)

    @needs_program('file')
    @needs_codec(program, 'xz')
    def test_star_xz_file(self):
        """Run archive commands with renamed TAR XZ archive."""
        self.archive_commands('t.tar.xz.foo', skip_create=True)



#!/usr/bin/env python3
# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""This is the main command line interface.
It will be made available as a script called "patool" on installation by setup.py.
The script will call the main_cli() function below on start.

Usage:
patool
   [global-options]
   {extract|list|create|repack|diff|search|formats}
   [sub-command-options]
   <command-args>
"""

import sys
import argparse
from . import (
    extract_archive,
    list_archive,
    test_archive,
    create_archive,
    diff_archives,
    search_archive,
    repack_archive,
    list_formats,
)
from .util import PatoolError
from .log import log_error, log_internal_error
from .configuration import App


def run_extract(args: argparse.Namespace) -> int:
    """Extract files from archive(s)."""
    res = 0
    for archive in args.archive:
        try:
            extract_archive(
                archive,
                verbosity=args.verbosity,
                interactive=args.interactive,
                outdir=args.outdir,
                password=args.password,
                format=args.format,
            )
        except PatoolError as msg:
            log_error(f"error extracting {archive}: {msg}")
            res += 1
    return res


def run_list(args: argparse.Namespace) -> int:
    """List files in archive(s)."""
    res = 0
    for archive in args.archive:
        try:
            verbosity = args.verbosity
            # increase default verbosity since the listing output should be visible
            if verbosity == 0:
                verbosity = 1
            list_archive(
                archive,
                verbosity=verbosity,
                interactive=args.interactive,
                password=args.password,
                format=args.format,
            )
        except PatoolError as msg:
            log_error(f"error listing {archive}: {msg}")
            res += 1
    return res


def run_test(args: argparse.Namespace) -> int:
    """Test files in archive(s)."""
    res = 0
    for archive in args.archive:
        try:
            test_archive(
                archive,
                verbosity=args.verbosity,
                interactive=args.interactive,
                password=args.password,
                format=args.format,
            )
        except PatoolError as msg:
            log_error(f"error testing {archive}: {msg}")
            res += 1
    return res


def run_create(args: argparse.Namespace) -> int:
    """Create an archive from given files."""
    res = 0
    try:
        create_archive(
            args.archive,
            args.filename,
            verbosity=args.verbosity,
            interactive=args.interactive,
            password=args.password,
            format=args.format,
        )
    except PatoolError as msg:
        log_error(f"error creating {args.archive}: {msg}")
        res = 1
    return res


def run_diff(args: argparse.Namespace) -> int:
    """Show differences between two archives."""
    try:
        res = diff_archives(
            args.archive1,
            args.archive2,
            verbosity=args.verbosity,
            interactive=args.interactive,
        )
    except PatoolError as msg:
        log_error(
            f"error showing differences between {args.archive1} and {args.archive2}: {msg}"
        )
        res = 2
    return res


def run_search(args: argparse.Namespace) -> int:
    """Search for pattern in given archive."""
    try:
        res = search_archive(
            args.pattern,
            args.archive,
            verbosity=args.verbosity,
            interactive=args.interactive,
            password=args.password,
        )
    except PatoolError as msg:
        log_error(f"error searching {args.archive}: {msg}")
        res = 2
    return res


def run_repack(args: argparse.Namespace) -> int:
    """Repackage one archive in another format."""
    res = 0
    try:
        repack_archive(
            args.archive_src,
            args.archive_dst,
            verbosity=args.verbosity,
            interactive=args.interactive,
        )
    except PatoolError as msg:
        log_error(f"error repacking {args.archive_src}: {msg}")
        res = 1
    return res


def run_formats(args: argparse.Namespace) -> int:
    """List supported and available archive formats."""
    list_formats()
    return 0


def run_version(args: argparse.Namespace) -> int:
    """Print version number."""
    print(App)
    return 0


Examples = """\
EXAMPLES
  patool extract archive.zip otherarchive.rar
  patool --verbose test dist.tar.gz
  patool list package.deb
  patool --verbose create myfiles.zip file1.txt dir/
  patool diff release1.0.tar.xz release2.0.zip
  patool search "def urlopen" python-3.3.tar.gz
  patool repack linux-2.6.33.tar.gz linux-2.6.33.tar.bz2
"""

Version = f"""\
VERSION
  {App}
"""


def create_argparser() -> argparse.ArgumentParser:
    """Construct and return an argument parser."""
    epilog = Examples + "\n" + Version
    parser = argparse.ArgumentParser(
        description="An archive file manager.",
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--verbose',
        '-v',
        action='count',
        default=0,
        dest='verbosity',
        help="verbose operation; can be given multiple times",
    )
    group.add_argument(
        '--quiet',
        '-q',
        action='count',
        default=0,
        dest='quiet',
        help="quiet operation; if given twice suppresses error output from archive commands",
    )
    parser.add_argument(
        '--non-interactive',
        dest='interactive',
        default=True,
        action='store_false',
        help="don't query for user input (i.e. passwords or when overwriting duplicate files); use with care since overwriting files or ignoring passwords could be unintended",
    )
    subparsers = parser.add_subparsers(
        help='the archive command; type "patool COMMAND -h" for command-specific help',
        dest='command',
        required=True,
    )
    # extract
    parser_extract = subparsers.add_parser(
        'extract', help='extract one or more archives'
    )
    parser_extract.add_argument('--outdir', help="output directory to extract to")
    parser_extract.add_argument('--password', help="password for encrypted files")
    parser_extract.add_argument(
        '--format',
        help="archive format (e.g., 7z, tar, rar) - auto detected if not provided",
    )
    parser_extract.add_argument('archive', nargs='+', help="an archive file")
    # list
    parser_list = subparsers.add_parser(
        'list', help='list members or one or more archives'
    )
    parser_list.add_argument('--password', help="password for encrypted files")
    parser_list.add_argument(
        '--format',
        help="archive format (e.g., 7z, tar, rar) - auto detected if not provided",
    )
    parser_list.add_argument('archive', nargs='+', help="an archive file")
    # create
    parser_create = subparsers.add_parser('create', help='create an archive')
    parser_create.add_argument('--password', help="password to encrypt files")
    parser_create.add_argument(
        '--format',
        help="archive format (e.g., 7z, tar, rar) - auto detected if not provided",
    )
    parser_create.add_argument(
        'archive',
        help="the archive file; the file extension determines the archive program",
    )
    parser_create.add_argument(
        'filename',
        nargs='+',
        help="a file or directory to add to the archive; note that some archive programs do not support directories",
    )
    # test
    parser_test = subparsers.add_parser('test', help='test an archive')
    parser_test.add_argument('--password', help="password for encrypted files")
    parser_test.add_argument(
        '--format',
        help="archive format (e.g., 7z, tar, rar) - auto detected if not provided",
    )
    parser_test.add_argument('archive', nargs='+', help='an archive file')
    # repack
    parser_repack = subparsers.add_parser(
        'repack', help='repack an archive to a different format'
    )
    parser_repack.add_argument('archive_src', help='source archive file')
    parser_repack.add_argument('archive_dst', help='target archive file')
    # diff
    parser_diff = subparsers.add_parser(
        'diff', help='show differences between two archives'
    )
    parser_diff.add_argument('archive1', help='the first archive file')
    parser_diff.add_argument('archive2', help='the second archive file')
    # search
    parser_search = subparsers.add_parser(
        'search', help="search contents of archive members"
    )
    parser_search.add_argument('--password', help="password for encrypted files")
    parser_search.add_argument('pattern', help='the grep(1) compatible search pattern')
    parser_search.add_argument('archive', help='the archive file')
    # formats
    subparsers.add_parser('formats', help="show supported archive formats")
    # version
    subparsers.add_parser('version', help="print version information")
    # optional bash completion
    try:
        import argcomplete  # noqa: PLC0415

        argcomplete.autocomplete(parser)
    except ImportError:
        pass
    return parser


def main(args=None) -> int:
    """Parse options and execute commands."""
    res = 0
    try:
        argparser = create_argparser()
        pargs = argparser.parse_args(args=args)
        if pargs.quiet:
            pargs.verbosity = -pargs.quiet
        # run subcommand function
        res = globals()[f"run_{pargs.command}"](pargs)
    except KeyboardInterrupt:
        log_error("aborted")
        res = 1
    except Exception:
        log_internal_error()
        res = 2
    return res


def main_cli():
    """Run main() function."""
    sys.exit(main())



# Copyright (C) 2010-2024 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""MIME type detection functions."""

import functools
import os
import mimetypes
import subprocess
from collections.abc import Sequence
from . import ArchiveMimetypes, ArchiveCompressions
from .log import log_error, log_warning, log_info
from .util import find_program, backtick


# internal MIME database
mimedb: mimetypes.MimeTypes | None = None


def init_mimedb() -> None:
    """Initialize the internal MIME database."""
    global mimedb  # noqa PLW0603
    try:
        mimedb = mimetypes.MimeTypes(strict=False)
    except Exception as msg:
        log_error(f"could not initialize MIME database: {msg}")
        return
    add_mimedb_data(mimedb)


def add_mimedb_data(mimedb: mimetypes.MimeTypes) -> None:
    """Add missing encodings and mimetypes to MIME database."""
    mimedb.encodings_map['.bz2'] = 'bzip2'
    mimedb.encodings_map['.lzma'] = 'lzma'
    mimedb.encodings_map['.xz'] = 'xz'
    mimedb.encodings_map['.lz'] = 'lzip'
    mimedb.encodings_map[".zst"] = "zstd"
    mimedb.suffix_map['.tbz2'] = '.tar.bz2'
    add_mimetype(mimedb, 'application/x-lzop', '.lzo')
    add_mimetype(mimedb, 'application/x-adf', '.adf')
    add_mimetype(mimedb, 'application/x-arj', '.arj')
    add_mimetype(mimedb, 'application/x-bzip3', '.bz3')
    add_mimetype(mimedb, 'application/x-lzma', '.lzma')
    add_mimetype(mimedb, 'application/x-xz', '.xz')
    add_mimetype(mimedb, 'application/java-archive', '.jar')
    add_mimetype(mimedb, 'application/vnd.rar', '.rar')
    add_mimetype(mimedb, 'application/vnd.rar', '.cbr')
    add_mimetype(mimedb, 'application/x-7z-compressed', '.7z')
    add_mimetype(mimedb, 'application/x-7z-compressed', '.cb7')
    add_mimetype(mimedb, 'application/x-cab', '.cab')
    add_mimetype(mimedb, 'application/x-rpm', '.rpm')
    add_mimetype(mimedb, 'application/x-debian-package', '.deb')
    add_mimetype(mimedb, 'application/x-ace', '.ace')
    add_mimetype(mimedb, 'application/x-ace', '.cba')
    add_mimetype(mimedb, 'application/x-archive', '.a')
    add_mimetype(mimedb, 'application/x-alzip', '.alz')
    add_mimetype(mimedb, 'application/x-arc', '.arc')
    # note: application/x-freearc is not added to mimedb since it uses the same file extension .arc as application/x-arc
    add_mimetype(mimedb, 'application/x-lrzip', '.lrz')
    add_mimetype(mimedb, 'application/x-lha', '.lha')
    add_mimetype(mimedb, 'application/x-lzh', '.lzh')
    add_mimetype(mimedb, 'application/x-lz4', '.lz4')
    add_mimetype(mimedb, 'application/x-rzip', '.rz')
    add_mimetype(mimedb, 'application/x-zoo', '.zoo')
    add_mimetype(mimedb, 'application/x-dms', '.dms')
    add_mimetype(mimedb, 'application/x-ms-wim', '.wim')
    add_mimetype(mimedb, 'application/x-zip-compressed', '.crx')
    add_mimetype(mimedb, 'application/x-shar', '.shar')
    add_mimetype(mimedb, 'application/x-tar', '.cbt')
    add_mimetype(mimedb, 'application/x-vhd', '.vhd')
    add_mimetype(mimedb, 'audio/x-ape', '.ape')
    add_mimetype(mimedb, 'audio/x-shn', '.shn')
    add_mimetype(mimedb, 'audio/flac', '.flac')
    add_mimetype(mimedb, 'application/x-chm', '.chm')
    add_mimetype(mimedb, 'application/x-iso9660-image', '.iso')
    add_mimetype(mimedb, 'application/zip', '.cbz')
    add_mimetype(mimedb, 'application/zip', '.epub')
    add_mimetype(mimedb, 'application/zip', '.apk')
    add_mimetype(mimedb, 'application/zpaq', '.zpaq')
    add_mimetype(mimedb, "application/zstd", ".zst")


def add_mimetype(mimedb: mimetypes.MimeTypes, mimetype: str, extension: str) -> None:
    """Add or replace a mimetype to be used with the given extension."""
    # If extension is already a common type, strict=True must be used.
    strict = extension in mimedb.types_map[True]
    mimedb.add_type(
        mimetype, extension, strict=strict
    )  # pytype: disable=attribute-error


@functools.cache
def guess_mime(filename: str) -> tuple[str | None, str | None]:
    """Guess the MIME type of given filename using file(1) and if that
    fails by looking at the filename extension with the Python mimetypes
    module.

    The result of this function is cached.
    """
    mime, encoding = guess_mime_file(filename)
    if mime is None:
        # fall back to guessing archive type by file extension
        mime, encoding = guess_mime_mimedb(filename)
    else:
        # check if file extension detection differs.
        mime2, _encoding2 = guess_mime_mimedb(filename)
        if mime2 != mime:
            log_info(
                f"Different MIME types detected for {filename}: "
                f"{mime} by file(1), {mime2} by extension. Preferring {mime}."
            )
    assert mime is not None or encoding is None
    return mime, encoding


Encoding2Mime: dict[str, str] = {
    'gzip': "application/gzip",
    'bzip2': "application/x-bzip2",
    'compress': "application/x-compress",
    'lzma': "application/x-lzma",
    'lzip': "application/x-lzip",
    'xz': "application/x-xz",
    "zstd": "application/zstd",
}
Mime2Encoding: dict[str, str] = dict(
    [(_val, _key) for _key, _val in Encoding2Mime.items()]
)

LegacyMimeType: dict[str, str] = {
    # libmagic before version 5.14 identified .gz files as application/x-gzip
    'application/x-gzip': "application/gzip",
    # libmagic before version 5.46 identified .rar files as application/x-rar
    'application/x-rar': "application/vnd.rar",
}


def guess_mime_mimedb(filename: str) -> tuple[str | None, str | None]:
    """Guess MIME type from given filename.
    @return: tuple (mime, encoding)
    """
    mime, encoding = None, None
    if mimedb is not None:
        mime, encoding = mimedb.guess_type(filename, strict=False)
    if mimedb is not None and mime is None and encoding is None:
        # try with lowercase extension, since we configure our mimedb entries only with lowercase
        # this way, files like "t.GZ" are recognized
        root, ext = os.path.splitext(filename)
        mime, encoding = mimedb.guess_type(root + ext.lower(), strict=False)
    if mime not in ArchiveMimetypes and encoding in ArchiveCompressions:
        # Files like 't.txt.gz' are recognized with encoding as format, and
        # an unsupported mime-type like 'text/plain'. Fix this.
        mime = Encoding2Mime[encoding]
        encoding = None
    return mime, encoding


def guess_mime_file(filename: str) -> tuple[str | None, str | None]:
    """Determine MIME type of filename with file(1):
     (a) using `file --brief --mime-type`
     (b) using `file --brief` and look at the result string
     (c) detect compressed archives (eg. .tar.gz) using
         `file --brief --mime --uncompress --no-sandbox`
    @return: tuple (mime, encoding)
    """
    mime, encoding = None, None
    if os.path.isfile(filename):
        file_prog = find_program("file")
        if file_prog is not None:
            mime, encoding = guess_mime_file_mime(file_prog, filename)
            if mime is None:
                mime = guess_mime_file_text(file_prog, filename)
                encoding = None
        else:
            log_info(
                "could not find a 'file' executable, falling back to guess mime type by file extension"
            )
    if mime in Mime2Encoding:
        # try to look inside compressed archives
        mime2 = None
        if file_prog is not None:
            cmd = [
                file_prog,
                "--brief",
                "--mime",
                "--uncompress",
                "--no-sandbox",
                filename,
            ]
            try:
                outparts = backtick(cmd).strip().split(";")
                mime2 = outparts[0].split(" ", 1)[0]
            except (OSError, subprocess.CalledProcessError) as err:
                log_warning(f"error executing {cmd}: {err}")

        if mime2 in LegacyMimeType:
            mime2 = LegacyMimeType[mime2]

        # Some file(1) implementations return an empty or unknown mime type
        # when the uncompressor program is not installed, other
        # implementation return the original file type.
        # The following detects both cases.
        if (
            mime2 in ('application/x-empty', 'application/octet-stream')
            or mime2 in Mime2Encoding
            or not mime2
        ):
            # The uncompressor program file(1) uses is not installed
            # or is not able to uncompress.
            # Try to get mime information from the file extension.
            mime2, encoding2 = guess_mime_mimedb(filename)
            if mime2 in ArchiveMimetypes:
                mime = mime2
                encoding = encoding2
        elif mime2 in ArchiveMimetypes:
            mime = mime2
            encoding = get_file_mime_encoding(outparts)
    return mime, encoding


def guess_mime_file_mime(
    file_prog: str, filename: str
) -> tuple[str | None, str | None]:
    """Determine MIME type of filename with file(1) and --mime option.
    @return: tuple (mime, encoding)
    """
    mime, encoding = None, None
    cmd = [file_prog, "--brief", "--mime-type", filename]
    try:
        mime = backtick(cmd).strip()
    except OSError as err:
        # ignore errors, as file(1) is only a fallback
        log_warning(f"error executing {cmd}: {err}")

    if mime in LegacyMimeType:
        mime = LegacyMimeType[mime]

    if mime not in ArchiveMimetypes:
        mime, encoding = None, None
    return mime, encoding


def get_file_mime_encoding(parts: Sequence[str]) -> str | None:
    """Get encoding value from splitted output of file --mime --uncompress."""
    for part in parts:
        for subpart in part.split(" "):
            if subpart.startswith("compressed-encoding="):
                mime = subpart.split("=")[1].strip()
                return Mime2Encoding.get(mime)
    return None


# Match file(1) output text to mime types
FileText2Mime: dict[str, str] = {
    "7-zip archive data": "application/x-7z-compressed",
    "ACE archive data": "application/x-ace",
    "ALZ archive data": "application/x-alzip",
    "Amiga DOS disk": "application/x-adf",
    "ARJ archive data": "application/x-arj",
    "bzip2 compressed data": "application/x-bzip2",
    "bzip3 compressed data": "application/x-bzip3",
    "cpio archive": "application/x-cpio",
    "ASCII cpio archive": "application/x-cpio",
    "Debian binary package": "application/x-debian-package",
    "FreeArc archive": "application/x-freearc",
    "gzip compressed data": "application/gzip",
    "LZMA compressed data": "application/x-lzma",
    "LRZIP compressed data": "application/x-lrzip",
    "lzop compressed data": "application/x-lzop",
    "Microsoft Cabinet archive data": "application/vnd.ms-cab-compressed",
    "RAR archive data": "application/vnd.rar",
    "RPM ": "application/x-rpm",
    "POSIX tar archive": "application/x-tar",
    "xz compressed data": "application/x-xz",
    "Zip archive data": "application/zip",
    "compress'd data": "application/x-compress",
    "lzip compressed data": "application/x-lzip",
    "rzip compressed data": "application/x-rzip",
    "current ar archive": "application/x-archive",
    "LHa ": "application/x-lha",
    "ARC archive data": "application/x-arc",
    "Zoo archive data": "application/x-zoo",
    "DMS archive data": "application/x-dms",
    "Monkey's Audio": "audio/x-ape",
    "FLAC audio bitstream data": "audio/flac",
    "MS Windows HtmlHelp Data": "application/x-chm",
    "ZPAQ stream": "application/zpaq",
}


def guess_mime_file_text(file_prog: str, filename: str) -> str | None:
    """Determine MIME type of filename with file(1)."""
    cmd = [file_prog, "--brief", filename]
    try:
        output = backtick(cmd).strip()
    except OSError as err:
        # ignore errors, as file(1) is only a fallback
        log_warning(f"error executing {cmd}: {err}")
        return None
    # match output against known strings
    for matcher, mime in FileText2Mime.items():
        if output.startswith(matcher) and mime in ArchiveMimetypes:
            return mime
    return None


init_mimedb()



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the arc program."""

import os
from ..util import PatoolError


def _add_password_to_options(options, password):
    """Check password and add it to ARC options."""
    if password is None:
        return options
    if ' ' in password:
        raise PatoolError("Password for ARC can't contain spaces.")
    options += f'g{password}'
    return options


def extract_arc(
    archive, compression, cmd, verbosity, interactive, outdir, password=None
):
    """Extract a ARC archive."""
    # Since extracted files will be placed in the current directory,
    # the cwd argument has to be the output directory.
    options = _add_password_to_options('x', password)
    cmdlist = [cmd, options, os.path.abspath(archive)]
    return (cmdlist, {'cwd': outdir})


def list_arc(archive, compression, cmd, verbosity, interactive, password=None):
    """List a ARC archive."""
    cmdlist = [cmd]
    if verbosity > 1:
        cmdlist.append(_add_password_to_options('v', password))
    else:
        cmdlist.append(_add_password_to_options('l', password))
    cmdlist.append(archive)
    return cmdlist


def test_arc(archive, compression, cmd, verbosity, interactive, password=None):
    """Test a ARC archive."""
    return [cmd, _add_password_to_options('t', password), archive]


def create_arc(
    archive, compression, cmd, verbosity, interactive, filenames, password=None
):
    """Create a ARC archive."""
    cmdlist = [cmd, _add_password_to_options('a', password), archive]
    cmdlist.extend(filenames)
    return cmdlist


extract_freearc = extract_arc
list_freearc = list_arc
test_freearc = test_arc
create_freearc = create_arc



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the nomarch program."""

import os


def extract_arc(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract an ARC archive."""
    # Since extracted files will be placed in the current directory,
    # the cwd argument has to be the output directory.
    cmdlist = [cmd, os.path.abspath(archive)]
    return (cmdlist, {'cwd': outdir})


def list_arc(archive, compression, cmd, verbosity, interactive):
    """List an ARC archive."""
    cmdlist = [cmd, '-l']
    if verbosity > 1:
        cmdlist.append('-v')
    cmdlist.append(archive)
    return cmdlist


def test_arc(archive, compression, cmd, verbosity, interactive):
    """Test an ARC archive."""
    return [cmd, '-t', archive]



# Copyright (C) 2012-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the zipfile Python module."""

from .. import util
import zipfile
import os

READ_SIZE_BYTES = 1024 * 1024


def list_zip(archive, compression, cmd, verbosity, interactive, password=None):
    """List member of a ZIP archive with the zipfile Python module."""
    try:
        with zipfile.ZipFile(archive, "r") as zfile:
            if password:
                zfile.setpassword(pwd=password.encode())
            for name in zfile.namelist():
                if verbosity >= 0:
                    print(name)
    except Exception as err:
        raise util.PatoolError(f"error listing {archive}") from err
    return


test_zip = list_zip


def extract_zip(
    archive, compression, cmd, verbosity, interactive, outdir, password=None
):
    """Extract a ZIP archive with the zipfile Python module."""
    try:
        if password:
            password = password.encode()
        with zipfile.ZipFile(archive) as zfile:
            zfile.extractall(outdir, pwd=password)
    except Exception as err:
        raise util.PatoolError(f"error extracting {archive}") from err
    return


def create_zip(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a ZIP archive with the zipfile Python module."""
    try:
        with zipfile.ZipFile(archive, 'w') as zfile:
            for filename in filenames:
                if os.path.isdir(filename):
                    write_directory(zfile, filename)
                else:
                    zfile.write(filename)
    except Exception as err:
        raise util.PatoolError(f"error creating {archive}") from err
    return


def write_directory(zfile, directory):
    """Write recursively all directories and filenames to zipfile instance."""
    for dirpath, dirnames, filenames in os.walk(directory):  # noqa: B007
        zfile.write(dirpath)
        for filename in filenames:
            zfile.write(os.path.join(dirpath, filename))



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""A lot of these tests need an external compression program.
See https://salsa.debian.org/debian/patool/-/blob/master/debian/control
at the "Suggests:" for a list of packages with supported compression
programs.

The file type detection uses the file(1) program which uses a library
of "magic" patterns. The contents of the magic pattern library can vary
between distributions so the tests might not run on all systems.
"""

import os
import patoolib
import pytest
import importlib

basedir = os.path.dirname(__file__)
datadir = os.path.join(basedir, 'data')

# Python 3.x function name attribute
fnameattr = '__name__'


def _need_func(testfunc, name, description):
    """Decorator skipping test if given testfunc returns False."""

    def check_func(func):
        def newfunc(*args, **kwargs):
            if not testfunc(name):
                pytest.skip(f"{description} {name!r} is not available")
            return func(*args, **kwargs)

        setattr(newfunc, fnameattr, getattr(func, fnameattr))
        return newfunc

    return check_func


def needs_os(name):
    """Decorator skipping test if given operating system is not available."""
    return _need_func(lambda x: os.name == x, name, 'operating system')


def needs_program(name):
    """Decorator skipping test if given program is not available."""
    return _need_func(patoolib.util.find_program, name, 'program')


def needs_one_program(programs):
    """Decorator skipping test if not one of given programs are available."""
    return _need_func(
        lambda x: any(map(patoolib.util.find_program, x)), programs, 'programs'
    )


def needs_module(name):
    """Decorator skipping test if given module is not available."""

    def has_module(module):
        try:
            importlib.import_module(module)
            return True
        except ImportError:
            return False

    return _need_func(has_module, name, 'Python module')


def needs_codec(program, codec, commands=patoolib.ArchiveCommands):
    """Decorator skipping test if given program codec is not available."""

    def check_prog(f):
        def newfunc(*args, **kwargs):
            exe = patoolib.util.find_program(program)
            if not exe:
                pytest.skip(f"program `{program}' not available")
            for command in commands:
                if not has_codec(command, program, exe, codec):
                    pytest.skip(
                        f"codec `{codec}' for program `{program}' and command `{command}' not available"
                    )
            return f(*args, **kwargs)

        setattr(newfunc, fnameattr, getattr(f, fnameattr))
        return newfunc

    return check_prog


def has_codec(command, program, exe, codec):
    """Test if program supports given codec with given command."""
    if program in ('7z', '7zz', '7zzs', '7za'):
        if codec == 'rar':
            # 7zip can be optionally built without rar support
            # Notably on Debian, the non-free p7zip-rar package must be installed to support RAR for 7z
            return patoolib.util.p7zip_supports_rar(program)
        if codec == 'compress':
            return patoolib.util.p7zip_supports_compress(program)
    elif program == 'arc':
        return patoolib.util.get_arc_format(exe) == codec
    return patoolib.program_supports_compression(command, program, exe, codec)



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the lha program."""


def extract_lzh(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract a LZH archive."""
    opts = 'x'
    if verbosity > 1:
        opts += 'v'
    opts += f"w={outdir}"
    return [cmd, opts, archive]


def list_lzh(archive, compression, cmd, verbosity, interactive):
    """List a LZH archive."""
    cmdlist = [cmd]
    if verbosity > 1:
        cmdlist.append('v')
    else:
        cmdlist.append('l')
    cmdlist.append(archive)
    return cmdlist


def test_lzh(archive, compression, cmd, verbosity, interactive):
    """Test a LZH archive."""
    opts = 't'
    if verbosity > 1:
        opts += 'v'
    return [cmd, opts, archive]


def create_lzh(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a LZH archive."""
    opts = 'a'
    if verbosity > 1:
        opts += 'v'
    cmdlist = [cmd, opts, archive]
    cmdlist.extend(filenames)
    return cmdlist



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the gzip program"""

from . import ArchiveTest, Content
from .. import needs_program


class TestGzip(ArchiveTest):
    """Test class for the gzip program"""

    program = 'gzip'

    @needs_program(program)
    def test_gzip(self):
        """Run archive commands with GZIP and COMPRESS archive."""
        self.archive_commands('t.txt.gz', check=Content.Singlefile)
        self.archive_extract('t.txt.Z', check=Content.Singlefile)

    @needs_program('file')
    @needs_program(program)
    def test_gzip_file(self):
        """Run archive commands with renamed GZIP and COMPRESS archive."""
        self.archive_commands(
            't.txt.gz.foo', skip_create=True, check=Content.Singlefile
        )
        self.archive_extract('t.txt.Z.foo', check=Content.Singlefile)

    def get_expected_singlefile_output(self, archive):
        """Gzip restores the original filename for .gz files"""
        if archive.endswith(".Z.foo"):
            return "t.txt.Z"
        return "t.txt"



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the 7za program"""

from . import ArchiveTest, Content
from .. import needs_program, needs_codec


class Test7za(ArchiveTest):
    """Test class for the 7za program"""

    program = '7za'

    @needs_program(program)
    def test_7za(self):
        """Run archive commands with archives that 7za supports."""
        self.archive_commands('t .7z')
        self.archive_commands('t .cb7')
        self.archive_commands('t.zip')
        self.archive_commands('t.cbz')
        self.archive_list('t.txt.gz')
        self.archive_list('t.txt.bz2')
        self.archive_list('t.jar')
        self.archive_list('t.cab')
        self.archive_list('t.rpm')
        self.archive_extract('t.txt.gz', check=Content.Singlefile)
        self.archive_extract('t.txt.bz2', check=Content.Singlefile)
        self.archive_extract('t.jar', check=None)
        self.archive_extract('t.cab')
        self.archive_test('t.txt.gz')
        self.archive_test('t.txt.bz2')
        self.archive_test('t.jar')
        self.archive_test('t.cab')
        self.archive_create('t.txt.gz', check=Content.Singlefile)
        self.archive_create('t.txt.bz2', check=Content.Singlefile)

    @needs_program(program)
    @needs_codec(program, 'compress')
    def test_7za_compress(self):
        """Run archive commands with compress archives."""
        self.archive_list('t.txt.Z')
        self.archive_extract('t.txt.Z', check=Content.Singlefile)
        self.archive_test('t.txt.Z')

    @needs_program('file')
    @needs_program(program)
    def test_7za_file(self):
        """Run archive commands with renamed archives that 7za supports."""
        self.archive_commands('t.7z.foo', skip_create=True)
        self.archive_commands('t.cb7.foo', skip_create=True)
        self.archive_commands('t.zip.foo', skip_create=True)
        self.archive_commands('t.cbz.foo', skip_create=True)
        self.archive_list('t.txt.gz.foo')
        self.archive_list('t.txt.bz2.foo')
        self.archive_list('t.jar.foo')
        self.archive_list('t.cab.foo')
        self.archive_list('t.rpm.foo')
        self.archive_extract('t.txt.gz.foo', check=None)
        self.archive_extract('t.txt.bz2.foo', check=Content.Singlefile)
        self.archive_extract('t.jar.foo', check=None)
        self.archive_extract('t.cab.foo')
        self.archive_test('t.txt.gz.foo')
        self.archive_test('t.txt.bz2.foo')
        self.archive_test('t.jar.foo')
        self.archive_test('t.cab.foo')

    @needs_program('file')
    @needs_program(program)
    @needs_codec(program, 'compress')
    def test_7za_compress_file(self):
        """Run archive commands with renamed compress archives."""
        self.archive_list('t.txt.Z.foo')
        self.archive_extract('t.txt.Z.foo', check=Content.Singlefile)
        self.archive_test('t.txt.Z.foo')



# Copyright (C) 2013-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the isoinfo program."""


def list_iso(archive, compression, cmd, verbosity, interactive):
    """List an ISO archive."""
    # Use Joliet (-J) and Rock-Ridge (-R) options.
    return [cmd, '-l', '-R', '-J', '-i', archive]



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test mime detection"""

import unittest
import os
import patoolib
import patoolib.mime
from . import needs_program, datadir


class TestMime(unittest.TestCase):
    """Test class for mime detection."""

    def mime_test(self, func, filename, mime, encoding, info):
        """Test that file has given mime and encoding as determined by
        given function.
        """
        self.assertTrue(os.path.isfile(filename), f"archive {filename!r} is not a file")
        file_mime, file_encoding = func(filename)
        fail_msg = "%s for archive `%s' should be %s, but was %s. Info: %s"
        if isinstance(mime, tuple):
            self.assertTrue(
                file_mime in mime,
                fail_msg % ("MIME type", filename, f"in {mime}", file_mime, info),
            )
        else:
            self.assertEqual(
                file_mime,
                mime,
                fail_msg % ("MIME type", filename, mime, file_mime, info),
            )
        self.assertEqual(
            file_encoding,
            encoding,
            fail_msg % ("Encoding", filename, encoding, file_encoding, info),
        )

    def mime_test_file(self, filename, mime, encoding=None):
        """Test that file has given mime and encoding as determined by
        file(1).
        """
        archive = os.path.join(datadir, filename)
        file_prog = patoolib.util.find_program("file")
        if file_prog is None:
            self.assertTrue(False, "could not find program 'file' in path")
        else:
            info = patoolib.mime.guess_mime_file_text(file_prog, archive)
            self.mime_test(patoolib.mime.guess_mime_file, archive, mime, encoding, info)

    def mime_test_mimedb(self, filename, mime, encoding=None):
        """Test that file has given mime and encoding as determined by the
        mimetypes module.
        """
        archive = os.path.join(datadir, filename)
        self.mime_test(patoolib.mime.guess_mime_mimedb, archive, mime, encoding, "")

    @needs_program('file')
    def test_mime_file(self):
        """Test mime detection of different archives with file(1)"""
        self.mime_test_file("t .7z", "application/x-7z-compressed")
        self.mime_test_file("t .cb7", "application/x-7z-compressed")
        self.mime_test_file("t.cb7.foo", "application/x-7z-compressed")
        self.mime_test_file("t.arj", "application/x-arj")
        self.mime_test_file("t.arj.foo", "application/x-arj")
        self.mime_test_file("t.txt.bz2", "application/x-bzip2")
        self.mime_test_file("t.txt.bz2.foo", "application/x-bzip2")
        # fails on macos-15 github ci
        # self.mime_test_file("t.txt.bz3", "application/x-bzip3")
        # self.mime_test_file("t.txt.bz3.foo", "application/x-bzip3")
        self.mime_test_file("t.cab", "application/vnd.ms-cab-compressed")
        self.mime_test_file("t.cab.foo", "application/vnd.ms-cab-compressed")
        self.mime_test_file("t.cpio", "application/x-cpio")
        self.mime_test_file("t.cpio.foo", "application/x-cpio")
        self.mime_test_file("t.deb", "application/x-debian-package")
        self.mime_test_file("t.deb.foo", "application/x-debian-package")
        self.mime_test_file("t.txt.gz", "application/gzip")
        self.mime_test_file("t.txt.gz.foo", "application/gzip")
        self.mime_test_file("t.jar", "application/zip")
        self.mime_test_file("t.jar.foo", "application/zip")
        self.mime_test_file("t.txt.lzma", "application/x-lzma")
        self.mime_test_file("t.txt.lzma.foo", "application/x-lzma")
        self.mime_test_file("t.txt.lz", "application/x-lzip")
        self.mime_test_file("t.txt.lz.foo", "application/x-lzip")
        self.mime_test_file("t.txt.lzo", "application/x-lzop")
        self.mime_test_file("t.txt.lzo.foo", "application/x-lzop")
        self.mime_test_file("t.rar", "application/vnd.rar")
        self.mime_test_file("t.rar.foo", "application/vnd.rar")
        self.mime_test_file("t.cbr", "application/vnd.rar")
        self.mime_test_file("t.cbr.foo", "application/vnd.rar")
        self.mime_test_file("t.rpm", "application/x-rpm")
        self.mime_test_file("t.rpm.foo", "application/x-rpm")
        self.mime_test_file("t.tar", "application/x-tar")
        self.mime_test_file("t.tar.foo", "application/x-tar")
        self.mime_test_file("t.cbt", "application/x-tar")
        self.mime_test_file("t.cbt.foo", "application/x-tar")
        # fails on macos-15 and windows-2025 github ci
        # self.mime_test_file("t.tar.lz", "application/x-tar", "lzip")
        self.mime_test_file("t.tar.bz2", "application/x-tar", "bzip2")
        self.mime_test_file("t.tbz2", "application/x-tar", "bzip2")
        self.mime_test_file("t.tar.gz", "application/x-tar", "gzip")
        self.mime_test_file("t.taz", "application/x-tar", "gzip")
        self.mime_test_file("t.tgz", "application/x-tar", "gzip")
        # fails on windows-latest github ci
        # self.mime_test_file("t.tar.xz", "application/x-tar", "xz")
        # self.mime_test_file("t.tar.Z", "application/x-tar", "compress")
        # self.mime_test_file("t.tar.lzma", "application/x-tar", "lzma")
        # self.mime_test_file("t.tar.zst", "application/x-tar", "zstd")
        # file(1) cannot uncompress .lzma files
        # self.mime_test_file("t.tar.lzma.foo", "application/x-tar", "lzma")
        self.mime_test_file("t.txt.xz", "application/x-xz")
        self.mime_test_file("t.txt.xz.foo", "application/x-xz")
        self.mime_test_file("t.txt.Z", "application/x-compress")
        self.mime_test_file("t.txt.Z.foo", "application/x-compress")
        self.mime_test_file("t.txt.zst", "application/zstd")
        self.mime_test_file("t.txt.zst.foo", "application/zstd")
        self.mime_test_file("t.jar", "application/zip")
        self.mime_test_file("t.jar.foo", "application/zip")
        self.mime_test_file("t.zip", "application/zip")
        self.mime_test_file("t.zip.foo", "application/zip")
        self.mime_test_file("t.cbz", "application/zip")
        self.mime_test_file("t.cbz.foo", "application/zip")
        self.mime_test_file("t.ace", "application/x-ace")
        self.mime_test_file("t.ace.foo", "application/x-ace")
        self.mime_test_file("t.cba", "application/x-ace")
        self.mime_test_file("t.cba.foo", "application/x-ace")
        # test errors on windows
        # self.mime_test_file("t.txt.a", "application/x-archive")
        # self.mime_test_file("t.txt.a.foo", "application/x-archive")
        self.mime_test_file("t.lha", "application/x-lha")
        self.mime_test_file("t.lzh", "application/x-lha")
        self.mime_test_file("t.lha.foo", "application/x-lha")
        self.mime_test_file("t.alz", "application/x-alzip")
        self.mime_test_file("t.alz.foo", "application/x-alzip")
        self.mime_test_file("t.arc", "application/x-arc")
        self.mime_test_file("t.arc.foo", "application/x-arc")
        self.mime_test_file("t.freearc.arc", "application/x-freearc")
        self.mime_test_file("t.freearc.arc.foo", "application/x-freearc")
        self.mime_test_file("t.txt.lz4", "application/x-lz4")
        self.mime_test_file("t.txt.lz4.foo", "application/x-lz4")
        self.mime_test_file("t.txt.lrz", "application/x-lrzip")
        self.mime_test_file("t.txt.lrz.foo", "application/x-lrzip")
        self.mime_test_file("t.txt.rz", "application/x-rzip")
        self.mime_test_file("t.txt.rz.foo", "application/x-rzip")
        self.mime_test_file("t.zoo", "application/x-zoo")
        self.mime_test_file("t.zoo.foo", "application/x-zoo")
        self.mime_test_file("t.dms", "application/x-dms")
        self.mime_test_file("t.dms.foo", "application/x-dms")
        self.mime_test_file("t.ape", "audio/x-ape")
        self.mime_test_file("t.ape.foo", "audio/x-ape")
        # file(1) does not recognize .shn files
        # self.mime_test_file("t.shn", "audio/x-shn")
        # self.mime_test_file("t.shn.foo", "audio/x-shn")
        self.mime_test_file("t.flac", "audio/flac")
        self.mime_test_file("t.flac.foo", "audio/flac")
        self.mime_test_file("t.adf", "application/x-adf")
        self.mime_test_file("t.adf.foo", "application/x-adf")
        self.mime_test_file("t.chm", "application/x-chm")
        self.mime_test_file("t.chm.foo", "application/x-chm")
        self.mime_test_file("t.iso", "application/x-iso9660-image")
        self.mime_test_file("t.epub", "application/zip")
        self.mime_test_file(
            "t.apk",
            (
                "application/zip",
                "application/java-archive",
                "application/jar",
                "application/vnd.android.package-archive",
            ),
        )
        self.mime_test_file("t.wim", "application/x-ms-wim")
        self.mime_test_file("t.wim.foo", "application/x-ms-wim")
        self.mime_test_file("t.zpaq", "application/zpaq")
        self.mime_test_file("t.zpaq.foo", "application/zpaq")

    # file(1) does not recognize .tar.lz files (at least not with --uncompress)
    # @needs_program('file')
    # @needs_program('lzip')
    # def test_mime_file_lzip(self):
    #    self.mime_test_file("t.tar.lz.foo", "application/x-tar", "lzip")

    @needs_program('file')
    @needs_program('bzip2')
    def test_mime_file_bzip(self):
        """Test mime detection of TAR BZIP2 archives"""
        self.mime_test_file("t.tar.bz2.foo", "application/x-tar", "bzip2")
        self.mime_test_file("t.tbz2.foo", "application/x-tar", "bzip2")

    @needs_program('file')
    def test_nested_gzip(self):
        """Test mime detection of archives with double compression"""
        # We won't extract this with rar, as it doesn't support archives wrapped in gzip
        # compression, but we will recognize the archive as a gzip-wrapped rar-file
        self.mime_test_file("t.rar.gz", "application/vnd.rar", "gzip")
        self.mime_test_file("t.rar.gz.foo", "application/vnd.rar", "gzip")

    @needs_program('file')
    @needs_program('gzip')
    def test_mime_file_gzip(self):
        """Test mime detection of TAR GZIP archives"""
        self.mime_test_file("t.tar.gz.foo", "application/x-tar", "gzip")
        self.mime_test_file("t.taz.foo", "application/x-tar", "gzip")
        self.mime_test_file("t.tgz.foo", "application/x-tar", "gzip")

    # broken tests
    # @needs_program('file')
    # @needs_program('xz')
    # def test_mime_file_xzip(self):
    #    """Test mime detection of TAR XZ archives"""
    #    self.mime_test_file("t.tar.xz.foo", "application/x-tar", "xz")

    # file(1) does not recognize .tar.zst files (at least not with --uncompress)
    # @needs_program("file")
    # @needs_program("zstd")
    # def test_mime_file_zstd(self):
    #    """Test mime detection of TAR ZSTD archives"""
    #    self.mime_test_file("t.tar.zst.foo", "application/x-tar", "zstd")

    # file(1) does not recognize .tar.Z files (at least not with --uncompress)
    # @needs_program('file')
    # @needs_program('uncompress')
    # def test_mime_file_compress(self):
    #    """Test mime detection of TAR COMPRESS archives"""
    #    self.mime_test_file("t.tar.Z.foo", "application/x-tar", "compress")

    def test_mime_mimedb(self):
        """Test mime detection of different archives with mime DB"""
        self.mime_test_mimedb("t .7z", "application/x-7z-compressed")
        self.mime_test_mimedb("t .cb7", "application/x-7z-compressed")
        self.mime_test_mimedb("t.arj", "application/x-arj")
        self.mime_test_mimedb("t.cab", "application/x-cab")
        self.mime_test_mimedb("t.cbr", "application/vnd.rar")
        self.mime_test_mimedb("t.cpio", "application/x-cpio")
        self.mime_test_mimedb("t.deb", "application/x-debian-package")
        self.mime_test_mimedb("t.jar", "application/java-archive")
        self.mime_test_mimedb("t.txt.lzma", "application/x-lzma")
        self.mime_test_mimedb("t.txt.lz", "application/x-lzip")
        self.mime_test_mimedb("t.txt.lz4", "application/x-lz4")
        self.mime_test_mimedb("t.txt.lzo", "application/x-lzop")
        self.mime_test_mimedb("t.rar", "application/vnd.rar")
        self.mime_test_mimedb("t.rpm", "application/x-rpm")
        self.mime_test_mimedb("t.tar", "application/x-tar")
        self.mime_test_mimedb("t.cbt", "application/x-tar")
        self.mime_test_mimedb("t.tar.bz2", "application/x-tar", "bzip2")
        self.mime_test_mimedb("t.tar.gz", "application/x-tar", "gzip")
        self.mime_test_mimedb("t.tar.lzma", "application/x-tar", "lzma")
        self.mime_test_mimedb("t.tar.xz", "application/x-tar", "xz")
        self.mime_test_mimedb("t.tar.lz", "application/x-tar", "lzip")
        self.mime_test_mimedb("t.tar.Z", "application/x-tar", "compress")
        self.mime_test_mimedb("t.tar.zst", "application/x-tar", "zstd")
        self.mime_test_mimedb("t.taz", "application/x-tar", "gzip")
        self.mime_test_mimedb("t.tbz2", "application/x-tar", "bzip2")
        self.mime_test_mimedb("t.tgz", "application/x-tar", "gzip")
        self.mime_test_mimedb("t.txt.gz", "application/gzip")
        self.mime_test_mimedb("t.txt.bz2", "application/x-bzip2")
        self.mime_test_mimedb("t.txt.bz3", "application/x-bzip3")
        self.mime_test_mimedb("t.txt.zst", "application/zstd")
        self.mime_test_mimedb("t.txt.xz", "application/x-xz")
        self.mime_test_mimedb("t.txt.Z", "application/x-compress")
        self.mime_test_mimedb(
            "t.zip", ("application/zip", "application/x-zip-compressed")
        )
        self.mime_test_mimedb(
            "t.cbz", ("application/zip", "application/x-zip-compressed")
        )
        self.mime_test_mimedb("t.ace", "application/x-ace")
        self.mime_test_mimedb("t.cba", "application/x-ace")
        self.mime_test_mimedb("t.txt.a", "application/x-archive")
        self.mime_test_mimedb("t.lha", "application/x-lha")
        self.mime_test_mimedb("t.lzh", "application/x-lzh")
        self.mime_test_mimedb("t.alz", "application/x-alzip")
        self.mime_test_mimedb("t.arc", "application/x-arc")
        self.mime_test_mimedb("t.txt.lrz", "application/x-lrzip")
        self.mime_test_mimedb("t.txt.rz", "application/x-rzip")
        self.mime_test_mimedb("t.zoo", "application/x-zoo")
        self.mime_test_mimedb("t.dms", "application/x-dms")
        self.mime_test_mimedb("t.shar", "application/x-shar")
        self.mime_test_mimedb("t.ape", "audio/x-ape")
        self.mime_test_mimedb("t.shn", "audio/x-shn")
        self.mime_test_mimedb("t.flac", "audio/flac")
        self.mime_test_mimedb("t.adf", "application/x-adf")
        self.mime_test_mimedb("t.chm", "application/x-chm")
        self.mime_test_mimedb("t.iso", "application/x-iso9660-image")
        self.mime_test_mimedb("t.epub", "application/zip")
        self.mime_test_mimedb("t.apk", "application/zip")
        self.mime_test_mimedb("t.vhd", "application/x-vhd")
        self.mime_test_mimedb("t.wim", "application/x-ms-wim")
        self.mime_test_mimedb("t.zpaq", "application/zpaq")



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Basic archive test utility functions."""

import unittest
import os
import patoolib
from .. import basedir, datadir

# All text files have '42' as content.
TextFileContent = '42'


class Content:
    """The test archives have one of several set of content files.
    The different content file sets have each a constant defined
    by this class.
    """

    # Recursive archives for extraction have a text file in a directory:
    # t/t.txt
    # Recursive archives for creation have two text files in directories:
    # foo dir/t.txt
    # foo dir/bar/t.txt
    Recursive = 'recursive'

    # Singlefile archives for extraction have a text file t.txt
    # Singlefile archives for creation have a text file `foo .txt'
    Singlefile = 'singlefile'

    # Multifile archives for extraction have two text files: t.txt and t2.txt
    # Multifile archives for creation have two text files: foo .txt and foo2 .txt
    Multifile = 'multifile'


class ArchiveTest(unittest.TestCase):
    """Helper class for archive tests, handling one commandline program."""

    # set program to use for archive handling in subclass
    program = None
    # set password for test with password
    password = None
    # default archive basename to check
    filename = 't'
    # verbosity levels in decreasing order, so that failing tests have maximum verbosity
    verbosity_levels = (2, 1, 0, -1, -2)

    def archive_commands(self, filename, **kwargs):
        """Run archive commands list, test, extract and create.
        All keyword arguments are delegated to the create test function.
        """
        self.archive_list(filename)
        if not kwargs.get('skip_test'):
            self.archive_test(filename)
        self.archive_extract(filename, check=kwargs.get('check', Content.Recursive))
        if not kwargs.get('skip_create'):
            self.archive_create(filename, **kwargs)

    def archive_extract(self, filename, check=Content.Recursive):
        """Test archive extraction."""
        archive = os.path.join(datadir, filename)
        self.assertTrue(
            os.path.isabs(archive), f"archive path is not absolute: {archive!r}"
        )
        self._archive_extract(archive, check)
        # archive name relative to tmpdir
        relarchive = os.path.join("..", archive[len(basedir) + 1 :])
        self._archive_extract(relarchive, check, verbosity=1)

    def _archive_extract(self, archive, check, verbosity=0):
        """Extract an archive into a temporary directory and check its
        contents.
        """
        # create a temporary directory for extraction
        tmpdir = patoolib.fileutil.tmpdir(dir=basedir, prefix="test_")
        try:
            olddir = patoolib.fileutil.chdir(tmpdir)
            try:
                output = patoolib.extract_archive(
                    archive,
                    program=self.program,
                    verbosity=verbosity,
                    interactive=False,
                    password=self.password,
                )
                if check:
                    self.check_extracted_archive(archive, output, check)
            finally:
                if olddir:
                    os.chdir(olddir)
        finally:
            patoolib.fileutil.rmtree(tmpdir)

    def check_extracted_archive(self, archive, output, check):
        """Check contents of an extracted archive"""
        if check == Content.Recursive:
            # outdir is the 't' directory of the archive
            self.assertEqual(output, 't')
            self.check_directory(output, 't')
            txtfile = os.path.join(output, 't.txt')
            self.check_textfile(txtfile, 't.txt')
        elif check == Content.Singlefile:
            # a non-existing directory to ensure files do not exist in it
            expected_output = self.get_expected_singlefile_output(archive)
            self.check_textfile(output, expected_output)
        elif check == Content.Multifile:
            txtfile = os.path.join(output, 't.txt')
            self.check_textfile(txtfile, 't.txt')
            txtfile2 = os.path.join(output, 't2.txt')
            self.check_textfile(txtfile2, 't2.txt')

    def get_expected_singlefile_output(self, archive):
        """Return the expected output file name of a singlefile archive."""
        if archive.endswith(".foo"):
            return patoolib.fileutil.stripext(os.path.basename(archive))
        return "t.txt"

    def check_directory(self, dirname, expectedname):
        """Check that directory exists."""
        self.assertTrue(os.path.isdir(dirname), dirname)
        self.assertEqual(os.path.basename(dirname), expectedname)

    def check_textfile(self, filename, expectedname):
        """Check that filename exists and has the default content."""
        self.assertTrue(os.path.isfile(filename), repr(filename))
        self.assertEqual(os.path.basename(filename), expectedname)
        self.assertEqual(get_filecontent(filename), TextFileContent)

    def archive_list(self, filename):
        """Test archive listing."""
        archive = os.path.join(datadir, filename)
        for verbosity in self.verbosity_levels:
            patoolib.list_archive(
                archive,
                program=self.program,
                verbosity=verbosity,
                interactive=False,
                password=self.password,
            )

    def archive_test(self, filename):
        """Test archive testing."""
        archive = os.path.join(datadir, filename)
        for verbosity in self.verbosity_levels:
            patoolib.test_archive(
                archive,
                program=self.program,
                verbosity=verbosity,
                interactive=False,
                password=self.password,
            )

    def archive_create(self, archive, srcfiles=None, check=Content.Recursive):
        """Test archive creation."""
        if srcfiles is None:
            if check == Content.Recursive:
                srcfiles = ('t',)
            elif check == Content.Singlefile:
                srcfiles = ('t.txt',)
            elif check == Content.Multifile:
                srcfiles = (
                    't.txt',
                    't2.txt',
                )
            else:
                raise ValueError('invalid check value {check!r}')
        olddir = patoolib.fileutil.chdir(datadir)
        try:
            # The format and compression arguments are needed for creating
            # archives with unusual file extensions.
            for verbosity in self.verbosity_levels:
                self._archive_create(
                    archive, srcfiles, program=self.program, verbosity=verbosity
                )
        finally:
            if olddir:
                os.chdir(olddir)

    def _archive_create(self, archive, srcfiles, program=None, verbosity=0):
        """Create archive from filename."""
        for srcfile in srcfiles:
            self.assertFalse(os.path.isabs(srcfile))
            self.assertTrue(os.path.exists(srcfile))
        # create a temporary directory for creation
        tmpdir = patoolib.fileutil.tmpdir(dir=basedir)
        try:
            archive = os.path.join(tmpdir, archive)
            self.assertTrue(
                os.path.isabs(archive), f"archive path is not absolute: {archive!r}"
            )
            patoolib.create_archive(
                archive,
                srcfiles,
                verbosity=verbosity,
                interactive=False,
                program=program,
                password=self.password,
            )
            self.assertTrue(os.path.isfile(archive))
            self.check_created_archive_with_test(archive)
            self.check_created_archive_with_diff(archive, srcfiles)
        finally:
            patoolib.fileutil.rmtree(tmpdir)

    def check_created_archive_with_test(self, archive):
        """Check given archive filename. Some archives need different
        test programs than the programs who created the archives.
        """
        command = patoolib.test_archive
        program = self.program
        # special case for programs that cannot test what they create
        if self.program in ('compress', 'py_gzip', 'zopfli'):
            program = 'gzip'
        elif self.program == 'py_bz2':
            program = 'bzip2'
        elif self.program == 'py_lzma':
            program = 'xz'
        elif self.program == 'py_zstd':
            program = 'zstd'
        elif self.program == 'zip':
            program = 'unzip'
        elif self.program in ('rzip', 'shorten'):
            program = 'py_echo'
            command = patoolib.list_archive
        elif self.program == 'lcab':
            program = 'cabextract'
        elif self.program == 'genisoimage':
            program = '7z'
        elif self.program == 'shar':
            return
        command(archive, program=program, password=self.password)

    def check_created_archive_with_diff(self, archive, srcfiles):
        """Extract created archive again and compare the contents."""
        # diff srcfile and output
        diff = patoolib.util.find_program("diff")
        if not diff:
            return
        program = self.program
        # special case for programs that cannot extract what they create
        if self.program in ('compress', 'zopfli'):
            program = 'gzip'
        elif self.program == 'zip':
            program = 'unzip'
        elif self.program == 'lcab':
            program = 'cabextract'
        elif self.program == 'shar':
            program = 'unshar'
        elif self.program == 'genisoimage':
            program = '7z'
        tmpdir = patoolib.fileutil.tmpdir(dir=basedir)
        try:
            olddir = patoolib.fileutil.chdir(tmpdir)
            try:
                output = patoolib.extract_archive(
                    archive, program=program, interactive=False, password=self.password
                )
                if len(srcfiles) == 1:
                    source = os.path.join(datadir, srcfiles[0])
                    patoolib.util.run_checked([diff, "-urN", source, output])
                else:
                    for srcfile in srcfiles:
                        source = os.path.join(datadir, srcfile)
                        target = os.path.join(output, srcfile)
                        patoolib.util.run_checked([diff, "-urN", source, target])
            finally:
                if olddir:
                    os.chdir(olddir)
        finally:
            patoolib.fileutil.rmtree(tmpdir)


def get_filecontent(filename):
    """Get file data as text."""
    with open(filename) as fo:
        return fo.read()


def get_nonexisting_directory(basedir):
    """Note: this is _not_ intended to be used to create a directory."""
    return patoolib.fileutil.get_single_outfile(basedir, "foo")



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the python print function"""

from . import ArchiveTest
from .. import needs_program


class TestPyecho(ArchiveTest):
    """Test class for the python print function"""

    program = 'py_echo'

    def test_py_echo(self):
        """List different single-file archives."""
        self.archive_list('t.txt.bz2')
        self.archive_list('t.txt.Z')
        self.archive_list('t.txt.lzma')
        self.archive_list('t.txt.lz')
        self.archive_list('t.txt.lrz')
        self.archive_list('t.txt.rz')
        self.archive_list('t.ape')
        self.archive_list('t.shn')
        self.archive_list('t.flac')

    @needs_program('file')
    def test_py_echo_file(self):
        """List different renamed single-file archives."""
        self.archive_list('t.txt.bz2.foo')
        self.archive_list('t.txt.Z.foo')
        # file(1) does not recognize .lzma files
        # self.archive_list('t.lzma.foo')
        self.archive_list('t.txt.lz.foo')
        self.archive_list('t.txt.lrz')
        self.archive_list('t.txt.rz.foo')
        self.archive_list('t.ape.foo')
        # file(1) does not recognize .shn files
        # self.archive_list('t.shn.foo')
        self.archive_list('t.flac.foo')



# Copyright (C) 2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the unar program."""


def _extract(archive, compression, cmd, verbosity, interactive, outdir, password=None):
    """Extract an archive."""
    cmdlist = [cmd, '-o', outdir]
    if password:
        cmdlist.extend(['-p', password])
    cmdlist.append(archive)
    return cmdlist


extract_rar = _extract
extract_zip = _extract
extract_7z = _extract
extract_tar = _extract
extract_gzip = _extract
extract_bzip2 = _extract
extract_lzma = _extract
extract_xz = _extract
extract_cab = _extract
extract_iso = _extract
extract_cpio = _extract
extract_compress = _extract
extract_zoo = _extract
extract_lzh = _extract
extract_dms = _extract



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for utility functions"""

import unittest
import os
from patoolib import fileutil


class UtilTest(unittest.TestCase):
    """Test class for utility functions"""

    def test_samefile1(self):
        """Test is_same_file* with absolute paths"""
        filename1 = filename2 = __file__
        self.assertTrue(fileutil.is_same_filename(filename1, filename2))
        self.assertTrue(fileutil.is_same_file(filename1, filename2))

    def test_samefile2(self):
        """Test is_same_file* with relative paths"""
        parentdir = os.path.dirname(__file__)
        filename1 = os.path.dirname(parentdir)
        filename2 = os.path.join(parentdir, '..')
        self.assertTrue(fileutil.is_same_filename(filename1, filename2))
        self.assertTrue(fileutil.is_same_file(filename1, filename2))

    def test_samefile3(self):
        """Test is_same_file* with different files"""
        parentdir = os.path.dirname(__file__)
        filename1 = os.path.dirname(parentdir)
        filename2 = os.path.join(parentdir, '.')
        self.assertFalse(fileutil.is_same_file(filename1, filename2))
        self.assertFalse(fileutil.is_same_filename(filename1, filename2))

    def test_stripext(self):
        """Test stripext() with different filenames"""
        self.assertTrue(fileutil.stripext("bar.gz") == "bar")
        self.assertTrue(fileutil.stripext("foo/bar.tar.gz") == "bar")
        self.assertTrue(fileutil.stripext("foo/bartar.gz") == "bartar")
        self.assertTrue(fileutil.stripext("foo/bar.7z") == "bar")
        self.assertTrue(fileutil.stripext("foo/bar") == "bar")

    def test_rmtree(self):
        """Test rmtree() with non-existing and temporary dirs"""
        parentdir = os.path.dirname(__file__)
        nonexisting_dir = os.path.join(parentdir, "imadoofus")
        fileutil.rmtree(nonexisting_dir)
        existing_dir = fileutil.tmpdir()
        self.assertTrue(os.path.exists(existing_dir))
        with open(os.path.join(existing_dir, "t.txt"), "w") as f:
            f.write("42")
        fileutil.rmtree(existing_dir)
        self.assertFalse(os.path.exists(existing_dir))

    def test_is_within_directory(self):
        """Test is_within_directory()"""
        curdir = os.getcwd()
        dest = os.path.join(curdir, "outputdir")
        member = os.path.join("outputdir2", "test.txt")
        self.assertFalse(fileutil.is_within_directory(dest, member))
        member = os.path.join("outputdir", "test.txt")
        self.assertTrue(fileutil.is_within_directory(dest, member))



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test patool diff command."""

import unittest
import os
from patoolib import cli
from . import datadir, needs_program


class ArchiveDiffTest(unittest.TestCase):
    """Test class for patool diff command."""

    @needs_program('diff')
    @needs_program('tar')
    @needs_program('unzip')
    def test_diff(self):
        """Run cli function to compare a TAR and ZIP archive."""
        archive1 = os.path.join(datadir, "t.tar")
        archive2 = os.path.join(datadir, "t.zip")
        args = ["-vv", "--non-interactive", "diff", archive1, archive2]
        cli.main(args=args)



# Copyright (C) 2012-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the lzma Python module.
Raises ImportError when neither compression.zstd nor pyzst module is found.
"""

from .. import fileutil, util, log

# try importing a python zstd module
try:
    from compression import zstd
except ImportError:
    import pyzstd as zstd  # ty: ignore[unresolved-import]

READ_SIZE_BYTES = 1024 * 1024


def extract_zstd(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract a ZSTD archive with the zstd Python module."""
    targetname = fileutil.get_single_outfile(outdir, archive)
    try:
        with zstd.ZstdFile(archive) as zstdfile:
            with open(targetname, 'wb') as targetfile:
                if verbosity >= 1:
                    log.log_info(f"extracting ZstdFile({archive}) to {targetname}")
                data = zstdfile.read(READ_SIZE_BYTES)
                while data:
                    targetfile.write(data)
                    data = zstdfile.read(READ_SIZE_BYTES)
    except Exception as err:
        msg = f"error extracting {archive} to {targetname}"
        raise util.PatoolError(msg) from err
    return


def create_zstd(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a ZSTD archive with the zstd Python module."""
    if len(filenames) > 1:
        raise util.PatoolError('multi-file compression not supported in Python zstd')
    try:
        with zstd.ZstdFile(archive, mode='wb') as zstdfile:
            filename = filenames[0]
            with open(filename, 'rb') as srcfile:
                if verbosity >= 1:
                    log.log_info("compressing {filename} to ZstdFile({archive})")
                data = srcfile.read(READ_SIZE_BYTES)
                while data:
                    zstdfile.write(data)
                    data = srcfile.read(READ_SIZE_BYTES)
    except Exception as err:
        raise util.PatoolError(f"error creating {archive}") from err
    return



# Copyright (C) 2013-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test patool create command."""

import unittest
import os
from patoolib import cli, fileutil
from . import basedir, datadir, needs_program


class ArchiveCreateTest(unittest.TestCase):
    """Test class for patool create command."""

    @needs_program('7z')
    def test_create(self):
        """Run cli function to create a 7Z archive."""
        tmpdir = fileutil.tmpdir(dir=basedir)
        try:
            files = [os.path.join(datadir, "t"), os.path.join(datadir, "t.txt")]
            archive = os.path.join(tmpdir, "t.7z")
            args = ["-vv", "--non-interactive", "create", archive]
            args.extend(files)
            cli.main(args=args)
        finally:
            fileutil.rmtree(tmpdir)



# Copyright (C) 2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the zstd program."""

from .. import util
from . import extract_singlefile_standard, test_singlefile_standard

extract_zstd = extract_singlefile_standard
test_zstd = test_singlefile_standard


def list_zstd(archive, compression, cmd, verbosity, interactive):
    """List a zstandard archive."""
    cmdlist = [cmd]
    cmdlist.append("-l")
    if verbosity > 1:
        cmdlist.append("-v")
    cmdlist.append(archive)
    return cmdlist


def create_zstd(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a zstandard archive."""
    cmdlist = [util.shell_quote(cmd)]
    if verbosity > 1:
        cmdlist.append("-v")
    cmdlist.extend(["-z", "--stdout", "--"])
    cmdlist.extend([util.shell_quote(x) for x in filenames])
    cmdlist.extend([">", util.shell_quote(archive)])
    return (cmdlist, {"shell": True})



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the gzip program."""

from . import extract_singlefile_standard, test_singlefile_standard
from .. import util, fileutil, log
import os
import subprocess

test_gzip = test_compress = test_singlefile_standard


def extract_gzip(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract given gzip archive."""
    cmdlist = [util.shell_quote(cmd)]
    if verbosity > 1:
        cmdlist.append('-v')
    outfile = get_original_filename(cmd, outdir, archive)
    cmdlist.extend(
        ['-c', '-d', '--', util.shell_quote(archive), '>', util.shell_quote(outfile)]
    )
    return (cmdlist, {'shell': True})


extract_compress = extract_singlefile_standard


def get_original_filename(cmd, outdir, archive):
    """Get the original filename inside a gzip archive.
    This parses the output of gzip --name --list.
    """
    outfile = fileutil.get_single_outfile(outdir, archive)
    cmdlist = [cmd, "--name", "--list", archive]
    try:
        output = util.backtick(cmdlist).strip()
        lines = output.splitlines()
        values = lines[1].split()
        baseoutfile = os.path.basename(outfile)
        basefilename = os.path.basename(values[-1])
        if os.path.normcase(baseoutfile) != os.path.normcase(basefilename):
            basearchive = os.path.basename(archive)
            # avoid overwriting the original archive with the same name
            if os.path.normcase(basefilename) != os.path.normcase(basearchive):
                outfile = os.path.join(outdir, basefilename)
    except (OSError, subprocess.CalledProcessError) as err:
        log.log_error(f"could not run {cmd}: {err}")
    return outfile


def create_gzip(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a GZIP archive."""
    cmdlist = [util.shell_quote(cmd)]
    if verbosity > 1:
        cmdlist.append('-v')
    cmdlist.extend(['-c', '--'])
    cmdlist.extend([util.shell_quote(x) for x in filenames])
    cmdlist.extend(['>', util.shell_quote(archive)])
    return (cmdlist, {'shell': True})


def list_gzip(archive, compression, cmd, verbosity, interactive):
    """List a GZIP archive."""
    cmdlist = [cmd]
    if verbosity > 0:
        cmdlist.append('-v')
    cmdlist.extend(['-l', '--', archive])
    return cmdlist



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the dpkg-deb program"""

from . import ArchiveTest
from .. import needs_program


class TestDpkg(ArchiveTest):
    """Test class for the dpkg-deb program"""

    program = 'dpkg-deb'

    @needs_program(program)
    def test_dpkg(self):
        """List, extract and test a DEB archive."""
        self.archive_list('t.deb')
        self.archive_extract('t.deb', check=None)
        self.archive_test('t.deb')

    @needs_program('file')
    @needs_program(program)
    def test_dpkg_file(self):
        """List, extract and test a renamed DEB archive."""
        self.archive_list('t.deb.foo')
        self.archive_extract('t.deb.foo', check=None)
        self.archive_test('t.deb.foo')



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Logging functions."""

import os
import sys
import time
import locale
import logging
import platform
from . import configuration


# the global logging.Logger object, initialized by init_logging()
logger: logging.Logger | None = None
default_encoding = locale.getpreferredencoding()


def init_logging(stream=sys.stderr) -> None:
    """Initialize the global logger. All log messages will be
    sent to the given stream, default is sys.stderr.
    """
    global logger  # noqa PLW0603
    logger = logging.getLogger(configuration.AppName)
    # do not propagate log message to higher log level handlers (in our case the root level),
    # since patoolib handles those messages by printing them on the given stream
    logger.propagate = False
    handler = logging.StreamHandler(stream=stream)
    format = f"%(levelname)s {configuration.AppName}: %(message)s"
    handler.setFormatter(logging.Formatter(format))
    logger.addHandler(handler)
    # the log level is set to INFO, and not changed any more since the verbosity parameter handles
    # log message printing instead (increased verbosity prints more messages).
    logger.setLevel(logging.INFO)


def encode_safe(*args, encoding=default_encoding) -> str:
    """Replacing unknown characters in args for the given encoding.
    @return: a space-separated string that will not have encoding errors
    with the given encoding
    """
    return " ".join(
        [str(arg).encode(encoding, errors="replace").decode(encoding) for arg in args]
    )


def log_error(msg) -> None:
    """Log error message."""
    if logger is not None:
        logger.error(encode_safe(msg))


def log_warning(msg) -> None:
    """Log warning message."""
    if logger is not None:
        logger.warning(encode_safe(msg))


def log_info(msg) -> None:
    """Log info message."""
    if logger is not None:
        logger.info(encode_safe(msg))


# environment keys to print for internal error info
EnvKeys = ("LANGUAGE", "LC_ALL", "LC_CTYPE", "LANG")


def log_internal_error() -> None:
    """Print internal error message."""
    now = time.localtime()
    env = os.linesep.join(
        [f"{key}={os.getenv(key)!r}" for key in EnvKeys if os.getenv(key) is not None]
    )
    if logger is not None:
        logger.exception(
            encode_safe(
                f"""********** Oops, I did it again. *************

You have found an internal error in {configuration.AppName}.
Please write a bug report at
{configuration.SupportUrl}
and include at least the information below:

Not disclosing some of the information below due to privacy reasons is ok.
I will try to help you nonetheless, but you have to give me something
I can work with ;) .

{configuration.App}
Python {sys.version} on {sys.platform}
Platform: {platform.platform()}
Local time: {strtime(now)}
sys.orig_argv: {sys.orig_argv}
Environment:
{env}
******** {configuration.AppName} internal error, over and out ********
"""
            )
        )


def strtime(t: time.struct_time) -> str:
    """Return ISO 8601 formatted time."""
    return time.strftime("%Y-%m-%d %H:%M:%S%z", t)


init_logging()



# Copyright (C) 2012-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the archmage program."""

import os
from .. import fileutil


def extract_chm(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract a CHM archive."""
    # archmage can only extract in non-existing directories
    # so a nice dirname is created
    name = fileutil.get_single_outfile("", archive)
    outfile = os.path.join(outdir, name)
    return [cmd, '-x', os.path.abspath(archive), outfile]


def test_chm(archive, compression, cmd, verbosity, interactive):
    """Test a CHM archive."""
    return [cmd, '-d', os.path.abspath(archive)]



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test patool configuration values"""

import unittest
import patoolib
import patoolib.mime


class TestConfiguration(unittest.TestCase):
    """Test class for patool configuration values"""

    def test_archive_mimetypes(self):
        """Test that each format has a MIME type"""
        self.assertEqual(
            set(patoolib.ArchiveFormats), set(patoolib.ArchiveMimetypes.values())
        )

    def test_archive_programs(self):
        """Test that the archive program key is an archive format"""
        self.assertEqual(
            set(patoolib.ArchiveFormats), set(patoolib.ArchivePrograms.keys())
        )
        for commands in patoolib.ArchivePrograms.values():
            for command in commands:
                if command is not None:
                    self.assertTrue(command in patoolib.ArchiveCommands)
            for programs in commands.values():
                self.assertTrue(isinstance(programs, tuple))

    def test_compression_programs(self):
        """Test that compressions are archive formats"""
        self.assertTrue(
            set(patoolib.ArchiveCompressions).issubset(set(patoolib.ArchiveFormats))
        )

    def test_encoding_mimes_keys(self):
        """Test that all compressions have a mime type"""
        self.assertEqual(
            set(patoolib.ArchiveCompressions), set(patoolib.mime.Encoding2Mime.keys())
        )

    def test_encoding_mimes_values(self):
        """Test that all encoding mimes are known mime types"""
        for mime in patoolib.mime.Encoding2Mime.values():
            self.assertTrue(mime in patoolib.ArchiveMimetypes)

    def test_filetext_mime(self):
        """Test that all text mime values are known mime types"""
        for mime in patoolib.mime.FileText2Mime.values():
            self.assertTrue(mime in patoolib.ArchiveMimetypes)



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Standard archive functions for singlefile archives."""

from collections.abc import Sequence
from .. import fileutil, util


def extract_singlefile_standard(
    archive: str,
    compression: str | None,
    cmd: str,
    verbosity: int,
    interactive: bool,
    outdir: str,
) -> tuple[Sequence[str], dict[str, bool]]:
    """Standard routine to extract a singlefile archive (like gzip)."""
    cmdlist = [util.shell_quote(cmd)]
    if verbosity > 1:
        cmdlist.append('-v')
    outfile = fileutil.get_single_outfile(outdir, archive)
    cmdlist.extend(
        ['-c', '-d', '--', util.shell_quote(archive), '>', util.shell_quote(outfile)]
    )
    return (cmdlist, {'shell': True})


def test_singlefile_standard(
    archive: str, compression: str | None, cmd: str, verbosity: int, interactive: bool
) -> Sequence[str]:
    """Standard routine to test a singlefile archive (like gzip)."""
    cmdlist = [cmd]
    if verbosity > 1:
        cmdlist.append('-v')
    cmdlist.extend(['-t', '--', archive])
    return cmdlist


def create_singlefile_standard(
    archive: str,
    compression: str | None,
    cmd: str,
    verbosity: int,
    interactive: bool,
    filenames: Sequence[str],
) -> tuple[Sequence[str], dict[str, bool]]:
    """Standard routine to create a singlefile archive (like gzip)."""
    cmdlist = [util.shell_quote(cmd)]
    if verbosity > 1:
        cmdlist.append('-v')
    cmdlist.extend(['-c', '--'])
    cmdlist.extend([util.shell_quote(x) for x in filenames])
    cmdlist.extend(['>', util.shell_quote(archive)])
    return (cmdlist, {'shell': True})



# Copyright (C) 2012-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the tarfile Python module."""

from .. import util, fileutil
import os
import sys
import tarfile


def list_tar(archive, compression, cmd, verbosity, interactive):
    """List a TAR archive with the tarfile Python module."""
    try:
        with tarfile.open(archive) as tfile:
            tfile.list(verbose=verbosity > 1)
    except Exception as err:
        raise util.PatoolError(f"error listing {archive}") from err
    return


test_tar = list_tar


def extract_tar(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract a TAR archive with the tarfile Python module."""
    try:
        with tarfile.open(archive) as tfile:
            if sys.version_info >= (3, 12, 0, "final", 0):
                tfile.extractall(path=outdir, filter='data')
            else:
                safe_extract(tfile, outdir)
    except Exception as err:
        raise util.PatoolError(f"error extracting {archive}") from err
    return


def safe_extract(tfile, path):
    """Helper function to ensure that TAR members will be extracted inside
    the given path.
    If a member will be extracted outside the path an Exception is raised.
    """
    safe_members = []
    bad_members = []
    for member in tfile.getmembers():
        member_path = os.path.join(path, member.name)
        if fileutil.is_within_directory(path, member_path):
            safe_members.append(member)
        else:
            bad_members.append(member)
    tfile.extractall(path, safe_members)
    if bad_members:
        filelist = ", ".join(member.name for member in bad_members)
        raise Exception(f"Unsafe tarfile entries: {filelist}.")


def create_tar(archive, compression, cmd, verbosity, interactive, filenames):
    """Create a TAR archive with the tarfile Python module."""
    mode = get_tar_mode(compression)
    try:
        with tarfile.open(archive, mode) as tfile:
            for filename in filenames:
                tfile.add(filename)
    except Exception as err:
        raise util.PatoolError(f"error creating {archive}") from err
    return


def get_tar_mode(compression):
    """Determine tarfile open mode according to the given compression."""
    if compression == 'gzip':
        return 'w:gz'
    if compression == 'bzip2':
        return 'w:bz2'
    if compression in ('lzma', 'xz'):
        return 'w:xz'
    if compression:
        msg = f'pytarfile does not support {compression} for tar compression'
        raise util.PatoolError(msg)
    # no compression
    return 'w'



# Copyright (C) 2025 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the FreeArc arc program"""

from . import ArchiveTest, Content
from .. import needs_program, needs_codec


class TestFreeArc(ArchiveTest):
    """Test class for the FreeArc arc program"""

    program = 'arc'

    @needs_program(program)
    @needs_codec(program, 'freearc')
    def test_freearc(self):
        """Run archive commands with FREEARC archive."""
        self.archive_commands(
            self.filename + '.freearc.arc', check=Content.Multifile, skip_create=True
        )



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands printing resulting filenames for single-file
compressed archives, implemented by the Python print function.
"""

from .. import fileutil


def list_bzip2(archive, compression, cmd, verbosity, interactive):
    """List a BZIP2 archive."""
    return stripext(cmd, archive, verbosity)


list_compress = list_lzma = list_xz = list_lzip = list_lrzip = list_rzip = list_bzip2


def list_ape(archive, compression, cmd, verbosity, interactive):
    """List an APE archive."""
    return stripext(cmd, archive, verbosity, extension=".wav")


list_shn = list_flac = list_ape


def stripext(cmd, archive, verbosity, extension=""):
    """Print the name without suffix."""
    if verbosity >= 0:
        print(fileutil.stripext(archive) + extension)
    return



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the rpm2cpio program."""

import os
from .. import util


def extract_rpm(archive, compression, cmd, verbosity, interactive, outdir):
    """Extract a RPM archive."""
    # also check cpio
    cpio = util.find_program("cpio")
    if not cpio:
        raise util.PatoolError(
            "cpio(1) is required for rpm2cpio extraction; please install it"
        )
    path = util.shell_quote(os.path.abspath(archive))
    cmdlist = [
        util.shell_quote(cmd),
        path,
        "|",
        util.shell_quote(cpio),
        '--extract',
        '--make-directories',
        '--preserve-modification-time',
        '--no-absolute-filenames',
        '--force-local',
        '--nonmatching',
        r'"*\.\.*"',
    ]
    if verbosity > 1:
        cmdlist.append('-v')
    return (cmdlist, {'cwd': outdir, 'shell': True})



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the arj program."""

from ..util import PatoolError


def _get_password_switch(password):
    """Check password and return password switch for ARJ."""
    if ' ' in password:
        raise PatoolError("Password for ARJ can't contain spaces.")
    return f'-g{password}'


def _maybe_add_password(cmdlist, password):
    if password:
        cmdlist.append(_get_password_switch(password))


def extract_arj(
    archive, compression, cmd, verbosity, interactive, outdir, password=None
):
    """Extract an ARJ archive."""
    cmdlist = [cmd, 'x', '-r']
    _maybe_add_password(cmdlist, password)
    if not interactive:
        cmdlist.append('-y')
    cmdlist.extend([archive, outdir])
    return cmdlist


def list_arj(archive, compression, cmd, verbosity, interactive, password=None):
    """List an ARJ archive."""
    cmdlist = [cmd]
    _maybe_add_password(cmdlist, password)
    if verbosity > 1:
        cmdlist.append('v')
    else:
        cmdlist.append('l')
    if not interactive:
        cmdlist.append('-y')
    cmdlist.extend(['-r', archive])
    return cmdlist


def test_arj(archive, compression, cmd, verbosity, interactive, password=None):
    """Test an ARJ archive."""
    cmdlist = [cmd, 't', '-r']
    _maybe_add_password(cmdlist, password)
    if not interactive:
        cmdlist.append('-y')
    cmdlist.append(archive)
    return cmdlist


def create_arj(
    archive, compression, cmd, verbosity, interactive, filenames, password=None
):
    """Create an ARJ archive."""
    cmdlist = [cmd, 'a', '-r']
    _maybe_add_password(cmdlist, password)
    if not interactive:
        cmdlist.append('-y')
    cmdlist.append(archive)
    cmdlist.extend(filenames)
    return cmdlist



# Copyright (C) 2010-2023 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Archive commands for the unace program."""

import os


def extract_ace(
    archive, compression, cmd, verbosity, interactive, outdir, password=None
):
    """Extract an ACE archive."""
    cmdlist = [cmd, 'x']
    if not outdir.endswith(os.sep):
        outdir += os.sep
    if password:
        cmdlist.append(f'-p{password}')
    cmdlist.extend([archive, outdir])
    return cmdlist


def list_ace(archive, compression, cmd, verbosity, interactive, password=None):
    """List an ACE archive."""
    cmdlist = [cmd]
    if verbosity > 1:
        cmdlist.append('v')
    else:
        cmdlist.append('l')
    if password:
        cmdlist.append(f'-p{password}')
    cmdlist.append(archive)
    return cmdlist


def test_ace(archive, compression, cmd, verbosity, interactive, password=None):
    """Test an ACE archive."""
    cmdlist = [cmd, 't']
    if password:
        cmdlist.append(f'-p{password}')
    cmdlist.append(archive)
    return cmdlist
